

"""
验证 LinterManager 配置文件自动加载功能的脚本

这个脚本演示了如何使用新的配置文件自动加载功能，
并测试不同的配置文件位置和格式。
"""

import os
import json
import yaml
import tempfile
import shutil
from pathlib import Path

from autocoder.common.linter_core.linter_manager import LinterManager


def create_test_config_directories():
    """创建测试配置目录结构"""
    print("=== 创建测试配置目录结构 ===")
    
    # 创建临时目录作为项目根目录
    project_root = tempfile.mkdtemp(prefix="linter_test_")
    print(f"项目根目录: {project_root}")
    
    # 创建不同优先级的配置目录
    config_dirs = {
        "project_level": os.path.join(project_root, ".autocoderlinters"),
        "project_auto_coder": os.path.join(project_root, ".auto-coder", ".autocoderlinters"),
        "global_level": os.path.expanduser("~/.auto-coder/.autocoderlinters")
    }
    
    for name, path in config_dirs.items():
        os.makedirs(path, exist_ok=True)
        print(f"创建目录: {name} -> {path}")
    
    return project_root, config_dirs


def create_sample_configs(config_dirs):
    """创建示例配置文件"""
    print("\n=== 创建示例配置文件 ===")
    
    # 项目级配置（最高优先级）
    project_config = {
        "max_workers": 8,
        "timeout": 600,
        "python_config": {
            "use_mypy": True,
            "flake8_args": ["--max-line-length=120"]
        },
        "source": "project_level"
    }
    
    project_config_file = os.path.join(config_dirs["project_level"], "config.json")
    with open(project_config_file, 'w') as f:
        json.dump(project_config, f, indent=2)
    print(f"创建项目级配置: {project_config_file}")
    
    # .auto-coder 级配置（中等优先级）
    auto_coder_config = {
        "max_workers": 6,
        "timeout": 500,
        "typescript_config": {
            "use_eslint": True,
            "tsc_args": ["--strict"]
        },
        "source": "auto_coder_level"
    }
    
    auto_coder_config_file = os.path.join(config_dirs["project_auto_coder"], "config.yaml")
    with open(auto_coder_config_file, 'w') as f:
        yaml.dump(auto_coder_config, f)
    print(f"创建 .auto-coder 级配置: {auto_coder_config_file}")
    
    # 全局配置（最低优先级）
    global_config = {
        "max_workers": 4,
        "timeout": 300,
        "java_config": {
            "javac_args": ["-Xlint:all"]
        },
        "source": "global_level"
    }
    
    global_config_file = os.path.join(config_dirs["global_level"], "config.yaml")
    with open(global_config_file, 'w') as f:
        yaml.dump(global_config, f)
    print(f"创建全局配置: {global_config_file}")
    
    return {
        "project": project_config,
        "auto_coder": auto_coder_config,
        "global": global_config
    }


def test_config_loading_priority(project_root, config_dirs, sample_configs):
    """测试配置加载优先级"""
    print("\n=== 测试配置加载优先级 ===")
    
    # 保存当前工作目录
    original_cwd = os.getcwd()
    
    try:
        # 切换到项目目录
        os.chdir(project_root)
        
        # 测试1: 所有配置文件都存在，应该加载项目级配置
        print("\n1. 测试所有配置文件都存在的情况:")
        manager = LinterManager()
        print(f"   加载的配置源: {manager.global_config.get('source', 'unknown')}")
        print(f"   max_workers: {manager.max_workers}")
        print(f"   timeout: {manager.timeout}")
        assert manager.global_config['source'] == 'project_level', "应该加载项目级配置"
        
        # 测试2: 删除项目级配置，应该加载 .auto-coder 级配置
        print("\n2. 测试删除项目级配置后:")
        os.remove(os.path.join(config_dirs["project_level"], "config.json"))
        manager = LinterManager()
        print(f"   加载的配置源: {manager.global_config.get('source', 'unknown')}")
        print(f"   max_workers: {manager.max_workers}")
        print(f"   timeout: {manager.timeout}")
        assert manager.global_config['source'] == 'auto_coder_level', "应该加载 .auto-coder 级配置"
        
        # 测试3: 删除 .auto-coder 级配置，应该加载全局配置
        print("\n3. 测试删除 .auto-coder 级配置后:")
        os.remove(os.path.join(config_dirs["project_auto_coder"], "config.yaml"))
        manager = LinterManager()
        print(f"   加载的配置源: {manager.global_config.get('source', 'unknown')}")
        print(f"   max_workers: {manager.max_workers}")
        print(f"   timeout: {manager.timeout}")
        assert manager.global_config['source'] == 'global_level', "应该加载全局配置"
        
        # 测试4: 删除所有配置文件，应该使用默认配置
        print("\n4. 测试删除所有配置文件后:")
        os.remove(os.path.join(config_dirs["global_level"], "config.yaml"))
        manager = LinterManager()
        print(f"   配置为空: {manager.global_config == {}}")
        print(f"   max_workers (默认): {manager.max_workers}")
        print(f"   timeout (默认): {manager.timeout}")
        assert manager.global_config == {}, "应该使用空配置"
        assert manager.max_workers == 4, "应该使用默认 max_workers"
        assert manager.timeout == 300, "应该使用默认 timeout"
        
    finally:
        # 恢复原始工作目录
        os.chdir(original_cwd)


def test_different_config_formats(project_root):
    """测试不同的配置文件格式"""
    print("\n=== 测试不同的配置文件格式 ===")
    
    original_cwd = os.getcwd()
    
    try:
        os.chdir(project_root)
        
        # 创建配置目录
        config_dir = os.path.join(project_root, ".autocoderlinters")
        os.makedirs(config_dir, exist_ok=True)
        
        # 测试 JSON 格式
        print("\n1. 测试 JSON 格式:")
        json_config = {
            "max_workers": 10,
            "format": "json",
            "python_config": {"use_mypy": True}
        }
        
        json_file = os.path.join(config_dir, "config.json")
        with open(json_file, 'w') as f:
            json.dump(json_config, f)
        
        manager = LinterManager()
        print(f"   格式: {manager.global_config.get('format')}")
        print(f"   max_workers: {manager.max_workers}")
        assert manager.global_config['format'] == 'json'
        
        # 清理并测试 YAML 格式
        os.remove(json_file)
        
        print("\n2. 测试 YAML 格式:")
        yaml_config = {
            "max_workers": 12,
            "format": "yaml",
            "typescript_config": {"use_eslint": True}
        }
        
        yaml_file = os.path.join(config_dir, "config.yaml")
        with open(yaml_file, 'w') as f:
            yaml.dump(yaml_config, f)
        
        manager = LinterManager()
        print(f"   格式: {manager.global_config.get('format')}")
        print(f"   max_workers: {manager.max_workers}")
        assert manager.global_config['format'] == 'yaml'
        
        # 清理并测试 YML 格式
        os.remove(yaml_file)
        
        print("\n3. 测试 YML 格式:")
        yml_config = {
            "max_workers": 14,
            "format": "yml",
            "java_config": {"javac_args": ["-Xlint:all"]}
        }
        
        yml_file = os.path.join(config_dir, "config.yml")
        with open(yml_file, 'w') as f:
            yaml.dump(yml_config, f)
        
        manager = LinterManager()
        print(f"   格式: {manager.global_config.get('format')}")
        print(f"   max_workers: {manager.max_workers}")
        assert manager.global_config['format'] == 'yml'
        
    finally:
        os.chdir(original_cwd)


def test_explicit_config_override():
    """测试显式配置覆盖自动加载"""
    print("\n=== 测试显式配置覆盖自动加载 ===")
    
    explicit_config = {
        "max_workers": 20,
        "timeout": 1000,
        "source": "explicit"
    }
    
    manager = LinterManager(explicit_config)
    print(f"配置源: {manager.global_config.get('source')}")
    print(f"max_workers: {manager.max_workers}")
    print(f"timeout: {manager.timeout}")
    
    assert manager.global_config['source'] == 'explicit'
    assert manager.max_workers == 20
    assert manager.timeout == 1000


def cleanup_test_directories(project_root, config_dirs):
    """清理测试目录"""
    print("\n=== 清理测试目录 ===")
    
    # 删除项目根目录
    if os.path.exists(project_root):
        shutil.rmtree(project_root)
        print(f"删除项目目录: {project_root}")
    
    # 删除全局配置目录（如果是我们创建的）
    global_dir = config_dirs["global_level"]
    if os.path.exists(global_dir):
        try:
            shutil.rmtree(global_dir)
            print(f"删除全局配置目录: {global_dir}")
        except OSError:
            print(f"无法删除全局配置目录: {global_dir} (可能有其他文件)")


def main():
    """主函数"""
    print("开始验证 LinterManager 配置文件自动加载功能")
    print("=" * 60)
    
    try:
        # 创建测试环境
        project_root, config_dirs = create_test_config_directories()
        sample_configs = create_sample_configs(config_dirs)
        
        # 运行测试
        test_config_loading_priority(project_root, config_dirs, sample_configs)
        test_different_config_formats(project_root)
        test_explicit_config_override()
        
        print("\n" + "=" * 60)
        print("✅ 所有测试通过！配置文件自动加载功能工作正常。")
        
    except Exception as e:
        print(f"\n❌ 测试失败: {e}")
        raise
    
    finally:
        # 清理测试环境
        if 'project_root' in locals() and 'config_dirs' in locals():
            cleanup_test_directories(project_root, config_dirs)
    
    print("\n验证完成！")


if __name__ == '__main__':
    main()




from typing import List, Union, Dict, Any, Optional
from pydantic import BaseModel
from autocoder.common.stream_out_type import IndexFilterStreamOutType
from autocoder.utils.auto_coder_utils.chat_stream_out import stream_out
from autocoder.common.utils_code_auto_generate import stream_chat_with_continue
from byzerllm.utils.str2model import to_model
from autocoder.index.types import IndexItem
from autocoder.common import AutoCoderArgs, SourceCode
import byzerllm
import time
from autocoder.index.index import IndexManager
from autocoder.index.types import (
    IndexItem,
    TargetFile,
    FileNumberList
)
from autocoder.common.tokens import count_string_tokens as count_tokens
from autocoder.common.printer import Printer
from concurrent.futures import ThreadPoolExecutor
from byzerllm import MetaHolder

from autocoder.utils.llms import get_llm_names, get_model_info
from loguru import logger
from byzerllm.utils.client.code_utils import extract_code
import json
from autocoder.index.symbols_utils import extract_symbols
import os.path
from autocoder.events.event_manager_singleton import get_event_manager
from autocoder.events import event_content as EventContentCreator
from autocoder.events.event_types import EventMetadata


def get_file_path(file_path):
    if file_path.startswith("##"):
        return file_path.strip()[2:]
    return file_path


class QuickFilterResult(BaseModel):
    files: Dict[str, TargetFile]
    has_error: bool
    error_message: Optional[str] = None
    file_positions: Optional[Dict[str, int]] = {}

    def get_sorted_file_positions(self) -> List[str]:
        """
        返回按 value 排序的文件列表
        """
        if not self.file_positions:
            return []
        return [file_path for file_path, _ in sorted(self.file_positions.items(), key=lambda x: x[1])]


class QuickFilter():
    def __init__(self, index_manager: IndexManager, stats: Dict[str, Any], sources: List[SourceCode]):
        self.index_manager = index_manager
        self.args = index_manager.args
        self.stats = stats
        self.sources = sources
        self.printer = Printer()
        self.max_tokens = self.args.index_filter_model_max_input_length

    def big_filter(self, index_items: List[IndexItem],) -> QuickFilterResult:
        chunks = []
        current_chunk = []

        # 将 index_items 切分成多个 chunks,第一个chunk尽可能接近max_tokens
        for item in index_items:
            # 使用 quick_filter_files.prompt 生成文本再统计
            temp_chunk = current_chunk + [item]
            prompt_text = self.quick_filter_files.prompt(
                temp_chunk, self.args.query)
            temp_size = count_tokens(prompt_text)
            # 如果当前chunk为空,或者添加item后不超过max_tokens,就添加到当前chunk
            if not current_chunk or temp_size <= self.max_tokens:
                current_chunk.append(item)
            else:
                # 当前chunk已满,创建新chunk
                chunks.append(current_chunk)
                current_chunk = [item]

        if current_chunk:
            chunks.append(current_chunk)

        tokens_len = count_tokens(
            self.quick_filter_files.prompt(index_items, self.args.query))
        self.printer.print_in_terminal(
            "quick_filter_too_long",
            style="yellow",
            tokens_len=tokens_len,
            max_tokens=self.max_tokens,
            split_size=len(chunks)
        )

        def process_chunk(chunk_index: int, chunk: List[IndexItem]) -> QuickFilterResult:
            try:
                # 获取模型名称列表
                model_names = get_llm_names(
                    self.index_manager.index_filter_llm)
                model_name = ",".join(model_names)
                files: Dict[str, TargetFile] = {}
                file_positions: Dict[str, int] = {}

                # 获取模型价格信息
                model_info_map = {}
                for name in model_names:
                    # 第二个参数是产品模式,从args中获取
                    info = get_model_info(name, self.args.product_mode)
                    if info:
                        model_info_map[name] = {
                            # 每百万tokens成本
                            "input_price": info.get("input_price", 0.0),
                            # 每百万tokens成本
                            "output_price": info.get("output_price", 0.0)
                        }

                if chunk_index == 0:
                    # 第一个chunk使用流式输出
                    stream_generator = stream_chat_with_continue(
                        self.index_manager.index_filter_llm,
                        [{"role": "user", "content": self.quick_filter_files.prompt(
                            chunk, self.args.query)}],
                        {},
                        self.args
                    )
                    full_response, last_meta = stream_out(
                        stream_generator,
                        model_name=model_name,
                        title=self.printer.get_message_from_key_with_format(
                            "quick_filter_title", model_name=model_name),
                        args=self.args,
                        extra_meta={
                            "stream_out_type": IndexFilterStreamOutType.FILE_NUMBER_LIST.value
                        }   
                    )
                    file_number_list = to_model(full_response, FileNumberList)

                    # 计算总成本
                    total_input_cost = 0.0
                    total_output_cost = 0.0

                    for name in model_names:
                        info = model_info_map.get(name, {})
                        # 计算公式:token数 * 单价 / 1000000
                        total_input_cost += (last_meta.input_tokens_count *
                                             info.get("input_price", 0.0)) / 1000000
                        total_output_cost += (last_meta.generated_tokens_count *
                                              info.get("output_price", 0.0)) / 1000000

                    # 四舍五入到4位小数
                    total_input_cost = round(total_input_cost, 4)
                    total_output_cost = round(total_output_cost, 4)

                    # 打印 token 统计信息和成本
                    self.printer.print_in_terminal(
                        "quick_filter_stats",
                        style="blue",
                        input_tokens=last_meta.input_tokens_count,
                        output_tokens=last_meta.generated_tokens_count,
                        input_cost=total_input_cost,
                        output_cost=total_output_cost,
                        model_names=model_name
                    )

                    get_event_manager(self.args.event_file).write_result(
                        EventContentCreator.create_result(content=EventContentCreator.ResultTokenStatContent(
                            model_name=model_name,
                            elapsed_time=0,
                            first_token_time=0,
                            input_tokens=last_meta.input_tokens_count,
                            output_tokens=last_meta.generated_tokens_count,
                            input_cost=total_input_cost,
                            output_cost=total_output_cost
                        ).to_dict()))
                else:
                    # 其他chunks直接使用with_llm
                    meta_holder = MetaHolder()
                    start_time = time.monotonic()                    
                    file_number_list = self.quick_filter_files.with_llm(self.index_manager.index_filter_llm).with_meta(
                        meta_holder).with_return_type(FileNumberList).run(chunk, self.args.query)
                    end_time = time.monotonic()
                    
                    total_input_cost = 0.0
                    total_output_cost = 0.0
                    if meta_holder.get_meta():
                        meta_dict = meta_holder.get_meta()                    
                        total_input_cost = meta_dict.get("input_tokens_count", 0) * model_info_map.get(model_name, {}).get("input_price", 0.0) / 1000000
                        total_output_cost = meta_dict.get("generated_tokens_count", 0) * model_info_map.get(model_name, {}).get("output_price", 0.0) / 1000000
                    
                    self.printer.print_in_terminal(
                        "quick_filter_stats",
                        style="blue",
                        input_tokens=meta_dict.get("input_tokens_count", 0),
                        output_tokens=meta_dict.get("generated_tokens_count", 0),
                        input_cost=total_input_cost,
                        output_cost=total_output_cost,
                        model_names=model_name,
                        elapsed_time=f"{end_time - start_time:.2f}"
                    )
                    get_event_manager(self.args.event_file).write_result(
                        EventContentCreator.create_result(content=EventContentCreator.ResultTokenStatContent(
                            model_name=model_name,
                            elapsed_time=end_time - start_time,
                            first_token_time=0,
                            input_tokens=meta_dict.get("input_tokens_count", 0),
                            output_tokens=meta_dict.get("generated_tokens_count", 0),
                            input_cost=total_input_cost,
                            output_cost=total_output_cost
                        ).to_dict()))

                if file_number_list:
                    for index,file_number in enumerate(file_number_list.file_list):
                        if file_number < 0 or file_number >= len(chunk):
                            self.printer.print_in_terminal(
                                "invalid_file_number",
                                style="yellow",
                                file_number=file_number,
                                total_files=len(chunk)
                            )
                            continue
                        file_path = get_file_path(
                            chunk[file_number].module_name)
                        files[file_path] = TargetFile(
                            file_path=chunk[file_number].module_name,
                            reason=self.printer.get_message_from_key(
                                "quick_filter_reason")
                        )
                        file_positions[file_path] = index
                return QuickFilterResult(
                    files=files,
                    has_error=False,
                    file_positions=file_positions
                )

            except Exception as e:
                self.printer.print_in_terminal(
                    "quick_filter_failed",
                    style="red",
                    error=str(e)
                )
                return QuickFilterResult(
                    files={},
                    has_error=True,
                    error_message=str(e)
                )

        results: List[QuickFilterResult] = []
        if chunks:
            with ThreadPoolExecutor() as executor:
                # 提交所有chunks到线程池并收集结果
                futures = [executor.submit(process_chunk, i, chunk)
                           for i, chunk in enumerate(chunks)]

                # 等待所有任务完成并收集结果
                for future in futures:
                    results.append(future.result())

        # 合并所有结果
        final_files: Dict[str, TargetFile] = {}
        final_file_positions: Dict[str, int] = {}
        has_error = False
        error_messages: List[str] = []

        for result in results:
            if result.has_error:
                has_error = True
                if result.error_message:
                    error_messages.append(result.error_message)
            final_files.update(result.files)


        for result in results:
            if result.has_error:
                has_error = True
                if result.error_message:
                    error_messages.append(result.error_message)
            ## 实现多个 result.file_positions 交织排序
            # 比如第一个是 {file_path_1_0: 0, file_path_1_1: 1, file_path_1_2: 2}    
            # 第二个是 {file_path_2_0: 0, file_path_2_1: 1}
            # 第三个是 {file_path_3_0: 0, file_path_3_1: 1, file_path_3_2: 2, file_path_3_3: 3}
            # 收集逻辑为所以 0 的为一组，然后序号为 0,1,2, 所有1 的为一组，序号是 3,4,5,依次往下推
            # {file_path_1_0: 0, file_path_2_0: 1, file_path_3_0: 2, file_path_1_1: 3, file_path_2_1: 4, file_path_3_1: 5}            
            # 
            # 获取所有结果的最大 position 值
            max_position = max([max(pos.values()) for pos in [result.file_positions for result in results if result.file_positions]] + [0])

            # 创建一个映射表，用于记录每个 position 对应的文件路径
            position_map = {}
            for result in results:
                if result.file_positions:
                    for file_path, position in result.file_positions.items():
                        if position not in position_map:
                            position_map[position] = []
                        position_map[position].append(file_path)

            # 重新排序文件路径
            new_file_positions = {}
            current_index = 0
            for position in range(max_position + 1):
                if position in position_map:
                    for file_path in position_map[position]:
                        new_file_positions[file_path] = current_index
                        current_index += 1

            # 更新 final_file_positions
            final_file_positions.update(new_file_positions)

        return QuickFilterResult(
            files=final_files,
            has_error=has_error,
            error_message="\n".join(error_messages) if error_messages else None
        )
    

    @byzerllm.prompt()
    def quick_filter_files(self, file_meta_list: List[IndexItem], query: str) -> str:
        '''
        当用户提一个需求的时候，我们要找到两种类型的源码文件：
        1. 根据需求需要被修改的文件，我们叫 edited_files
        2. 为了能够完成修改这些文件，还需要的一些额外参考文件, 我们叫 reference_files
        3. 因为修改了 edited_files 文件，可能有一些依赖 edited_files 的文件也需要被修改，我们叫 dependent_files

        现在，给定下面的索引文件：

        <index>
        {{ content }}
        </index>

        索引文件包含文件序号(##[]括起来的部分)，文件路径，文件符号信息等。            

        下面是用户的查询需求：

        <query>
        {{ query }}
        </query>

        请根据用户的需求，找到相关的文件，并给出文件序号列表。请返回如下json格式：

        ```json
        {
            "file_list": [
                file_index1,
                file_index2,
                ...
            ]
        }
        ```

        特别注意:    
        1. 如果用户的query里有 @文件 或者 @@符号，并且他们在索引文件中，那么被@的文件或者@@的符号必须要返回。
        2. 根据需求以及根据 @文件 或者 @@符号 找到的文件，猜测需要被修改的edited_files文件，然后尝试通过索引文件诸如导入语句等信息找到这些文件依赖的其他文件得到 reference_files,dependent_files。
        3. file_list 里的文件序号，按被 @ 或者 @@ 文件，edited_files文件，reference_files,dependent_files文件的顺序排列。注意，reference_files 你要根据需求来猜测是否需要，过滤掉不相关的，避免返回文件数过多。
        4. 如果 query 里是一段历史对话，那么对话里的内容提及的文件路径必须要返回。        
        5. 如果用户需求为空，则直接返回空列表即可。        
        6. 返回的 json格式数据不允许有注释
        '''

        file_meta_str = "\n".join(
            [f"##[{index}]{item.module_name}\n{item.symbols}" for index, item in enumerate(file_meta_list)])
        context = {
            "content": file_meta_str,
            "query": query            
        }
        return context
    

    def _extract_code_snippets_from_overflow_files(self, validated_file_numbers: List[int],index_items: List[IndexItem], conversations: List[Dict[str, str]]):
        token_count = 0        
        selected_files = []
        selected_file_contents = []
        full_file_tokens = int(self.max_tokens * 0.8)
        for file_number in validated_file_numbers:
            file_path = get_file_path(index_items[file_number].module_name)
            with open(file_path, "r", encoding="utf-8") as f:
                content = f.read()
            tokens = count_tokens(content)
            if token_count + tokens <= full_file_tokens:
                selected_files.append(file_number)
                selected_file_contents.append(content)
                token_count += tokens
            else:
                # 对超出部分抽取代码片段
                try:
                    extracted_info = (
                        self.extract_code_snippets_from_files.options(
                            {"llm_config": {"max_length": 100}}
                        )
                        .with_llm(self.index_manager.index_filter_llm)
                        .run(conversations, [content])
                    )
                    json_str = extract_code(extracted_info)[0][1]
                    json_objs = json.loads(json_str)

                    new_content = ""

                    if json_objs:                        
                        for json_obj in json_objs:
                            start_line = json_obj["start_line"] - 1
                            end_line = json_obj["end_line"]
                            chunk = "\n".join(content.split("\n")[start_line:end_line])
                            new_content += chunk + "\n"
                      
                        token_count += count_tokens(new_content)
                        if token_count >= self.max_tokens:
                            break
                        else:
                            selected_files.append(file_number)
                            selected_file_contents.append(new_content)
                except Exception as e:
                    logger.error(f"Failed to extract code snippets from {file_path}: {e}")
        return selected_files


    def filter(self, index_items: List[IndexItem], query: str) -> QuickFilterResult:
        final_files: Dict[str, TargetFile] = {}
        final_file_positions: Dict[str, int] = {}
        start_time = time.monotonic()

        prompt_str = self.quick_filter_files.prompt(index_items, query)

        tokens_len = count_tokens(prompt_str)

        # 打印当前索引大小
        self.printer.print_in_terminal(
            "quick_filter_tokens_len",
            style="blue",
            tokens_len=tokens_len
        )
        
        if tokens_len > self.max_tokens and tokens_len < 4*self.max_tokens:
            # 打印 big_filter 模式的状态
            self.printer.print_in_terminal(
                "filter_mode_big",
                style="yellow",
                tokens_len=tokens_len
            )
            return self.big_filter(index_items)
        elif tokens_len > 4*self.max_tokens:
            # 打印 super_big_filter 模式的状态
            self.printer.print_in_terminal(
                "filter_mode_super_big",
                style="yellow",
                tokens_len=tokens_len
            )
            round1 = self.super_big_filter(index_items)
            round1_index_items = []
            for file_path in round1.files.keys():
                for index_item in index_items:
                    if index_item.module_name == file_path:
                        round1_index_items.append(index_item)
                        
            if round1_index_items:
                round2 = self.big_filter(round1_index_items)
                return round2
            return round1
        else:
            # 打印普通过滤模式的状态
            self.printer.print_in_terminal(
                "filter_mode_normal",
                style="blue"
            )

        try:
            # 获取模型名称
            model_names = get_llm_names(self.index_manager.index_filter_llm)
            model_name = ",".join(model_names)

            # 获取模型价格信息
            model_info_map = {}
            for name in model_names:
                # 第二个参数是产品模式,从args中获取
                info = get_model_info(name, self.args.product_mode)
                if info:
                    model_info_map[name] = {
                        # 每百万tokens成本
                        "input_price": info.get("input_price", 0.0),
                        # 每百万tokens成本
                        "output_price": info.get("output_price", 0.0)
                    }

            # 渲染 Prompt 模板
            query = self.quick_filter_files.prompt(
                index_items, self.args.query)

            # 使用流式输出处理
            stream_generator = stream_chat_with_continue(
                self.index_manager.index_filter_llm,
                [{"role": "user", "content": query}],
                {},
                self.args
            )

            def extract_file_number_list(content: str) -> str:
                try:
                    v = to_model(content, FileNumberList)
                    return "\n".join([index_items[file_number].module_name for file_number in v.file_list])
                except Exception as e:
                    logger.error(f"Error extracting file number list: {e}")
                    return content

            # 获取完整响应
            full_response, last_meta = stream_out(
                stream_generator,
                model_name=model_name,
                title=self.printer.get_message_from_key_with_format(
                    "quick_filter_title", model_name=model_name),
                args=self.args,
                display_func=extract_file_number_list,
                extra_meta={
                    "stream_out_type": IndexFilterStreamOutType.FILE_NUMBER_LIST.value
                }
            )
            # 解析结果
            file_number_list = to_model(full_response, FileNumberList)
            end_time = time.monotonic()

            # 计算总成本
            total_input_cost = 0.0
            total_output_cost = 0.0

            for name in model_names:
                info = model_info_map.get(name, {})
                # 计算公式:token数 * 单价 / 1000000
                total_input_cost += (last_meta.input_tokens_count *
                                     info.get("input_price", 0.0)) / 1000000
                total_output_cost += (last_meta.generated_tokens_count *
                                      info.get("output_price", 0.0)) / 1000000

            # 四舍五入到4位小数
            total_input_cost = round(total_input_cost, 4)
            total_output_cost = round(total_output_cost, 4)
            speed = last_meta.generated_tokens_count / (end_time - start_time)

            # 打印 token 统计信息和成本
            self.printer.print_in_terminal(
                "quick_filter_stats",
                style="blue",
                elapsed_time=f"{end_time - start_time:.2f}",
                input_tokens=last_meta.input_tokens_count,
                output_tokens=last_meta.generated_tokens_count,
                input_cost=total_input_cost,
                output_cost=total_output_cost,
                model_names=model_name,
                speed=f"{speed:.2f}"
            )
            get_event_manager(self.args.event_file).write_result(
                EventContentCreator.create_result(content=EventContentCreator.ResultTokenStatContent(
                    model_name=model_name,
                    elapsed_time=end_time - start_time,
                    input_tokens=last_meta.input_tokens_count,
                    output_tokens=last_meta.generated_tokens_count,
                    input_cost=total_input_cost,
                    output_cost=total_output_cost,
                    speed=speed
                ).to_dict()))

        except Exception as e:
            self.printer.print_in_terminal(
                "quick_filter_failed",
                style="red",
                error=str(e)
            )
            return QuickFilterResult(
                files=final_files,
                has_error=True,
                error_message=str(e)
            )

        if file_number_list:
            validated_file_numbers = []
            for file_number in file_number_list.file_list:
                if file_number < 0 or file_number >= len(index_items):
                    self.printer.print_in_terminal(
                        "invalid_file_number",
                        style="yellow",
                        file_number=file_number,
                        total_files=len(index_items)
                    )
                    continue
                validated_file_numbers.append(file_number)
                                
            # 将最终选中的文件加入final_files
            for index,file_number in enumerate(validated_file_numbers):
                file_path = get_file_path(index_items[file_number].module_name)
                final_files[file_path] = TargetFile(
                    file_path=index_items[file_number].module_name,
                    reason=self.printer.get_message_from_key("quick_filter_reason")
                )
                final_file_positions[file_path] = index

        end_time = time.monotonic()
        self.stats["timings"]["quick_filter"] = end_time - start_time
        return QuickFilterResult(
            files=final_files,
            has_error=False,
            file_positions=final_file_positions
        )

    def super_big_filter(self, index_items: List[IndexItem]) -> QuickFilterResult:
        """
        超大索引过滤方法，通过提取文件的核心信息（文件名和用途）来减少token数量
        可处理超大规模的索引文件，通过切分成多个chunks并行处理
        """
        compact_items = []
        
        # 将每个索引项转换为更紧凑的格式：只保留文件名和用途
        for index, item in enumerate(index_items):
            # 从module_name中提取文件名
            filename = os.path.basename(item.module_name)
            
            # 从symbols中提取用途
            symbols_info = extract_symbols(item.symbols)
            usage = symbols_info.usage if symbols_info.usage else "无用途描述"
            
            # 创建紧凑的表示
            compact_item = {
                "index": index,
                "filename": filename,
                "full_path": item.module_name,
                "usage": usage
            }
            compact_items.append(compact_item)
        
        # 切分compact_items成多个chunks
        chunks = []
        current_chunk = []
        batch_size = 100  # 每100条记录检查一次token数量
        
        # 计算总的tokens长度
        full_prompt = self.super_big_quick_filter_files.prompt(compact_items, self.args.query)
        tokens_len = count_tokens(full_prompt)
        
        # 如果tokens长度不超过max_tokens，直接处理整个列表
        if tokens_len <= self.max_tokens:
            return self._process_compact_items(compact_items, index_items)
        
        # 否则，将compact_items切分成多个chunks，每100条检查一次
        for i, item in enumerate(compact_items):
            current_chunk.append(item)
            
            # 每处理batch_size条记录或者到达末尾时检查一次
            if (i + 1) % batch_size == 0 or i == len(compact_items) - 1:
                temp_prompt = self.super_big_quick_filter_files.prompt(current_chunk, self.args.query)
                temp_size = count_tokens(temp_prompt)
                
                # 如果当前chunk的token数超过限制，则从当前位置分割
                if temp_size > self.max_tokens:
                    # 如果当前chunk为空，添加至少一项
                    if len(current_chunk) <= batch_size:
                        # 当前批次是第一批，但已经超过限制，则至少保留一半的记录
                        split_index = max(1, len(current_chunk) // 2)
                        chunks.append(current_chunk[:split_index])
                        current_chunk = current_chunk[split_index:]
                    else:
                        # 将前一批次的items作为一个chunk
                        prev_batch_end = len(current_chunk) - (i % batch_size + 1)
                        if prev_batch_end > 0:
                            chunks.append(current_chunk[:prev_batch_end])
                            current_chunk = current_chunk[prev_batch_end:]
                        else:
                            # 极端情况：即使一条记录也超过了限制，则尝试添加单条记录
                            chunks.append([current_chunk[0]])
                            current_chunk = current_chunk[1:]
        
        # 确保最后的chunk也被添加
        if current_chunk:
            chunks.append(current_chunk)
        
        # 打印切分信息
        self.printer.print_in_terminal(
            "super_big_filter_splitting",
            style="yellow",
            tokens_len=tokens_len,
            max_tokens=self.max_tokens,
            split_size=len(chunks)
        )
        
        # 定义处理单个chunk的函数
        def process_chunk(chunk_index: int, chunk: List[dict]) -> QuickFilterResult:
            # 为避免在所有chunk上都显示UI，只在第一个chunk上显示
            if chunk_index == 0:
                # 显示UI的处理方式
                return self._process_compact_items(chunk, index_items, show_ui=True, chunk_index=chunk_index)
            else:
                # 非UI显示的处理方式
                return self._process_compact_items(chunk, index_items, show_ui=False, chunk_index=chunk_index)
        
        # 使用ThreadPoolExecutor并行处理所有chunks
        results: List[QuickFilterResult] = []
        if chunks:
            with ThreadPoolExecutor() as executor:
                futures = [executor.submit(process_chunk, i, chunk) for i, chunk in enumerate(chunks)]
                for future in futures:
                    results.append(future.result())
        
        # 合并所有结果
        final_files: Dict[str, TargetFile] = {}
        final_file_positions: Dict[str, int] = {}
        has_error = False
        error_messages: List[str] = []
        
        # 收集所有文件和错误信息
        for result in results:
            if result.has_error:
                has_error = True
                if result.error_message:
                    error_messages.append(result.error_message)
            final_files.update(result.files)
        
        # 处理file_positions的交织排序
        max_position = max([max(pos.values()) for pos in [result.file_positions for result in results if result.file_positions]] + [0])
        
        # 创建position映射表
        position_map = {}
        for result in results:
            if result.file_positions:
                for file_path, position in result.file_positions.items():
                    if position not in position_map:
                        position_map[position] = []
                    position_map[position].append(file_path)
        
        # 重新排序文件路径
        current_index = 0
        for position in range(max_position + 1):
            if position in position_map:
                for file_path in position_map[position]:
                    final_file_positions[file_path] = current_index
                    current_index += 1
        
        return QuickFilterResult(
            files=final_files,
            has_error=has_error,
            error_message="\n".join(error_messages) if error_messages else None,
            file_positions=final_file_positions
        )

    def _process_compact_items(self, compact_items: List[dict], index_items: List[IndexItem], show_ui: bool = True, chunk_index: int = 0) -> QuickFilterResult:
        """
        处理一组compact_items，返回QuickFilterResult
        """
        # 使用流式输出处理
        model_names = get_llm_names(self.index_manager.index_filter_llm)
        model_name = ",".join(model_names)
        
        # 获取模型价格信息
        model_info_map = {}
        for name in model_names:
            info = get_model_info(name, self.args.product_mode)
            if info:
                model_info_map[name] = {
                    "input_price": info.get("input_price", 0.0),
                    "output_price": info.get("output_price", 0.0)
                }
        
        try:
            start_time = time.monotonic()
            # 渲染 Prompt 模板
            prompt = self.super_big_quick_filter_files.prompt(compact_items, self.args.query)
            
            if show_ui:
                # 使用流式输出处理
                stream_generator = stream_chat_with_continue(
                    self.index_manager.index_filter_llm,
                    [{"role": "user", "content": prompt}],
                    {},
                    self.args
                )
                
                def extract_file_number_list(content: str) -> str:
                    try:
                        v = to_model(content, FileNumberList)
                        return "\n".join([index_items[compact_items[file_number]["index"]].module_name for file_number in v.file_list])
                    except Exception as e:
                        logger.error(f"Error extracting file number list: {e}")
                        return content
                
                # 获取完整响应
                full_response, last_meta = stream_out(
                    stream_generator,
                    model_name=model_name,
                    title=self.printer.get_message_from_key_with_format(
                        "super_big_filter_title", model_name=model_name),
                    args=self.args,
                    display_func=extract_file_number_list,
                    extra_meta={
                        "stream_out_type": IndexFilterStreamOutType.FILE_NUMBER_LIST.value
                    }
                )
                
                # 解析结果
                file_number_list = to_model(full_response, FileNumberList)
                end_time = time.monotonic()
                
                # 计算总成本
                total_input_cost = 0.0
                total_output_cost = 0.0
                
                for name in model_names:
                    info = model_info_map.get(name, {})
                    total_input_cost += (last_meta.input_tokens_count * 
                                        info.get("input_price", 0.0)) / 1000000
                    total_output_cost += (last_meta.generated_tokens_count * 
                                         info.get("output_price", 0.0)) / 1000000
                
                # 四舍五入到4位小数
                total_input_cost = round(total_input_cost, 4)
                total_output_cost = round(total_output_cost, 4)
                speed = last_meta.generated_tokens_count / (end_time - start_time)
                
                # 打印 token 统计信息和成本
                self.printer.print_in_terminal(
                    "super_big_filter_stats",
                    style="blue",
                    elapsed_time=f"{end_time - start_time:.2f}",
                    input_tokens=last_meta.input_tokens_count,
                    output_tokens=last_meta.generated_tokens_count,
                    input_cost=total_input_cost,
                    output_cost=total_output_cost,
                    model_names=model_name,
                    speed=f"{speed:.2f}",
                    chunk_index=chunk_index
                )

                get_event_manager(self.args.event_file).write_result(
                    EventContentCreator.create_result(content=EventContentCreator.ResultTokenStatContent(
                        model_name=model_name,
                        elapsed_time=end_time - start_time,
                        input_tokens=last_meta.input_tokens_count,
                        output_tokens=last_meta.generated_tokens_count,
                        input_cost=total_input_cost,
                        output_cost=total_output_cost,
                        speed=speed
                    ).to_dict()))
                
            else:
                # 非UI模式，直接使用LLM处理
                meta_holder = MetaHolder()
                file_number_list = self.super_big_quick_filter_files.with_llm(self.index_manager.index_filter_llm).with_meta(
                    meta_holder).with_return_type(FileNumberList).run(compact_items, self.args.query)
                end_time = time.monotonic()
                
                # 打印处理信息
                if meta_holder.get_meta():
                    meta_dict = meta_holder.get_meta()
                    total_input_cost = meta_dict.get("input_tokens_count", 0) * model_info_map.get(model_name, {}).get("input_price", 0.0) / 1000000
                    total_output_cost = meta_dict.get("generated_tokens_count", 0) * model_info_map.get(model_name, {}).get("output_price", 0.0) / 1000000
                    
                    self.printer.print_in_terminal(
                        "super_big_filter_stats",
                        style="blue",
                        input_tokens=meta_dict.get("input_tokens_count", 0),
                        output_tokens=meta_dict.get("generated_tokens_count", 0),
                        input_cost=total_input_cost,
                        output_cost=total_output_cost,
                        model_names=model_name,
                        elapsed_time=f"{end_time - start_time:.2f}",
                        chunk_index=chunk_index
                    )

                    get_event_manager(self.args.event_file).write_result(
                        EventContentCreator.create_result(content=EventContentCreator.ResultTokenStatContent(
                            model_name=model_name,
                            elapsed_time=end_time - start_time,
                            input_tokens=meta_dict.get("input_tokens_count", 0),
                            output_tokens=meta_dict.get("generated_tokens_count", 0),
                            input_cost=total_input_cost,
                            output_cost=total_output_cost,
                            speed=speed
                        ).to_dict()))
                    
            
            # 构建返回结果
            files = {}
            file_positions = {}
            
            if file_number_list:
                validated_file_numbers = []
                for file_number in file_number_list.file_list:
                    if file_number < 0 or file_number >= len(compact_items):
                        self.printer.print_in_terminal(
                            "invalid_file_number",
                            style="yellow",
                            file_number=file_number,
                            total_files=len(compact_items)
                        )
                        continue
                    
                    # 获取实际的index_item索引
                    original_index = compact_items[file_number]["index"]
                    validated_file_numbers.append(original_index)
                
                # 将最终选中的文件加入files
                for index, file_number in enumerate(validated_file_numbers):
                    file_path = get_file_path(index_items[file_number].module_name)
                    files[file_path] = TargetFile(
                        file_path=index_items[file_number].module_name,
                        reason=self.printer.get_message_from_key("quick_filter_reason")
                    )
                    file_positions[file_path] = index
            
            return QuickFilterResult(
                files=files,
                has_error=False,
                file_positions=file_positions
            )
        
        except Exception as e:
            self.printer.print_in_terminal(
                "super_big_filter_failed",
                style="red",
                error=str(e)
            )
            return QuickFilterResult(
                files={},
                has_error=True,
                error_message=str(e)
            )

    @byzerllm.prompt()
    def super_big_quick_filter_files(self, compact_items: List[dict], query: str) -> str:
        '''
        当用户提一个需求的时候，我们要找到相关的源码文件。
        
        下面是简化的索引文件列表，每项包含文件序号(index,##[]括起来的部分)、文件名(filename)和用途描述(usage)：

        <index>
        {{ file_meta_str }}        
        </index>

        下面是用户的查询需求：

        <query>
        {{ query }}
        </query>

        请根据用户的需求，找到相关的文件，并给出文件序号列表。请返回如下json格式：

        ```json
        {
            "file_list": [
                file_index1,
                file_index2,
                ...
            ]
        }
        ```

        特别注意:    
        1. 如果用户的query里有 @文件 或者 @@符号，请匹配对应的文件名，优先返回这些文件。
        2. 根据用户需求找出需要被修改的文件(edited_files)，以及可能需要作为参考的文件(reference_files)。
        3. file_list 里的文件序号，按被 @ 的文件、edited_files文件和reference_files文件的顺序排列。
        4. 如果 query 里是一段历史对话，那么对话里提及的文件必须要返回。
        5. 如果用户需求为空，则直接返回空列表即可。
        6. 返回的 json格式数据不允许有注释
        '''
        file_meta_str = "\n".join(
            [f"##[{index}]{item['filename']}\n{item['usage']}" for index, item in enumerate(compact_items)])
        
        context = {
            "file_meta_str": file_meta_str,
            "query": query
        }
        return context



import pytest
import os
import json
import tempfile
import shutil
from datetime import datetime, timedelta
from pathlib import Path

from autocoder.common.file_checkpoint.backup import FileBackupManager

@pytest.fixture
def temp_test_dir():
    """提供一个临时的测试目录"""
    temp_dir = tempfile.mkdtemp()
    yield temp_dir
    shutil.rmtree(temp_dir)

@pytest.fixture
def temp_backup_dir():
    """提供一个临时的备份目录"""
    temp_dir = tempfile.mkdtemp()
    yield temp_dir
    shutil.rmtree(temp_dir)

@pytest.fixture
def sample_file(temp_test_dir):
    """创建一个用于测试的样例文件"""
    file_path = os.path.join(temp_test_dir, "sample.txt")
    with open(file_path, 'w', encoding='utf-8') as f:
        f.write("这是一个测试文件的内容")
    return file_path

class TestFileBackupManager:
    """FileBackupManager类的单元测试"""
    
    def test_init_with_custom_dir(self, temp_backup_dir):
        """测试使用自定义目录初始化"""
        manager = FileBackupManager(backup_dir=temp_backup_dir)
        
        assert manager.backup_dir == temp_backup_dir
        assert os.path.exists(temp_backup_dir)
    

    def test_backup_file(self, temp_backup_dir, sample_file):
        """测试备份文件功能"""
        manager = FileBackupManager(backup_dir=temp_backup_dir)
        
        # 备份文件
        backup_id = manager.backup_file(sample_file)
        
        # 检查备份ID
        assert backup_id is not None
        assert len(backup_id) > 0
        
        # 检查备份文件是否存在
        backup_file_path = os.path.join(temp_backup_dir, backup_id)
        assert os.path.exists(backup_file_path)
        
        # 检查备份文件内容
        with open(backup_file_path, 'r', encoding='utf-8') as f:
            content = f.read()
        assert content == "这是一个测试文件的内容"
        
        # 检查元数据
        assert backup_id in manager.metadata
        assert manager.metadata[backup_id]["original_path"] == sample_file
        assert "timestamp" in manager.metadata[backup_id]
        assert "size" in manager.metadata[backup_id]
    
    def test_backup_nonexistent_file(self, temp_backup_dir):
        """测试备份不存在的文件"""
        manager = FileBackupManager(backup_dir=temp_backup_dir)
        
        # 尝试备份不存在的文件
        backup_id = manager.backup_file("nonexistent_file.txt")
        
        # 应该返回None
        assert backup_id is None
    
    def test_restore_file(self, temp_backup_dir, temp_test_dir, sample_file):
        """测试恢复文件功能"""
        manager = FileBackupManager(backup_dir=temp_backup_dir)
        
        # 备份文件
        backup_id = manager.backup_file(sample_file)
        
        # 修改原始文件
        with open(sample_file, 'w', encoding='utf-8') as f:
            f.write("已修改的内容")
        
        # 恢复到新位置
        restore_path = os.path.join(temp_test_dir, "restored.txt")
        success = manager.restore_file(restore_path, backup_id)
        
        # 检查恢复结果
        assert success is True
        assert os.path.exists(restore_path)
        
        # 检查恢复文件内容
        with open(restore_path, 'r', encoding='utf-8') as f:
            content = f.read()
        assert content == "这是一个测试文件的内容"
    
    def test_restore_with_invalid_backup_id(self, temp_backup_dir, temp_test_dir):
        """测试使用无效的备份ID恢复文件"""
        manager = FileBackupManager(backup_dir=temp_backup_dir)
        
        # 尝试恢复不存在的备份
        restore_path = os.path.join(temp_test_dir, "restored.txt")
        success = manager.restore_file(restore_path, "invalid_backup_id")
        
        # 应该失败
        assert success is False
        assert not os.path.exists(restore_path)
    
    def test_get_backup_content(self, temp_backup_dir, sample_file):
        """测试获取备份文件内容"""
        manager = FileBackupManager(backup_dir=temp_backup_dir)
        
        # 备份文件
        backup_id = manager.backup_file(sample_file)
        
        # 获取备份内容
        content = manager.get_backup_content(backup_id)
        
        # 检查内容
        assert content == "这是一个测试文件的内容"
    
    def test_get_backup_content_with_invalid_id(self, temp_backup_dir):
        """测试使用无效的备份ID获取内容"""
        manager = FileBackupManager(backup_dir=temp_backup_dir)
        
        # 尝试获取不存在的备份内容
        content = manager.get_backup_content("invalid_backup_id")
        
        # 应该返回None
        assert content is None
    
    def test_delete_backup(self, temp_backup_dir, sample_file):
        """测试删除备份"""
        manager = FileBackupManager(backup_dir=temp_backup_dir)
        
        # 备份文件
        backup_id = manager.backup_file(sample_file)
        backup_file_path = os.path.join(temp_backup_dir, backup_id)
        
        # 检查备份文件是否存在
        assert os.path.exists(backup_file_path)
        assert backup_id in manager.metadata
        
        # 删除备份
        success = manager.delete_backup(backup_id)
        
        # 检查删除结果
        assert success is True
        assert not os.path.exists(backup_file_path)
        assert backup_id not in manager.metadata
    
    def test_delete_nonexistent_backup(self, temp_backup_dir):
        """测试删除不存在的备份"""
        manager = FileBackupManager(backup_dir=temp_backup_dir)
        
        # 尝试删除不存在的备份
        success = manager.delete_backup("nonexistent_backup_id")
        
        # 应该返回False
        assert success is False
    
    def test_get_backups_for_file(self, temp_backup_dir, sample_file):
        """测试获取指定文件的所有备份"""
        manager = FileBackupManager(backup_dir=temp_backup_dir)
        
        # 创建多个备份
        backup_ids = []
        for i in range(3):
            backup_id = manager.backup_file(sample_file)
            backup_ids.append(backup_id)
        
        # 获取文件的备份列表
        backups = manager.get_backups_for_file(sample_file)
        
        # 检查备份列表
        assert len(backups) == 3
        for backup_id, timestamp in backups:
            assert backup_id in backup_ids
            assert isinstance(timestamp, float)
    
    @pytest.mark.parametrize("max_age_days", [1, 7, 30])
    def test_clean_old_backups(self, temp_backup_dir, sample_file, max_age_days):
        """测试清理旧备份"""
        manager = FileBackupManager(backup_dir=temp_backup_dir)
        
        # 创建一个备份
        backup_id = manager.backup_file(sample_file)
        
        # 修改备份的时间戳为过去的时间
        old_timestamp = (datetime.now() - timedelta(days=max_age_days+1)).timestamp()
        manager.metadata[backup_id]["timestamp"] = old_timestamp
        manager._save_metadata()
        
        # 创建一个新备份
        new_backup_id = manager.backup_file(sample_file)
        
        # 清理旧备份
        cleaned_count = manager.clean_old_backups(max_age_days)
        
        # 检查清理结果
        assert cleaned_count == 1
        assert backup_id not in manager.metadata
        assert not os.path.exists(os.path.join(temp_backup_dir, backup_id))
        assert new_backup_id in manager.metadata
        assert os.path.exists(os.path.join(temp_backup_dir, new_backup_id))
    
    def test_metadata_persistence(self, temp_backup_dir, sample_file):
        """测试元数据持久化"""
        # 创建一个备份
        manager1 = FileBackupManager(backup_dir=temp_backup_dir)
        backup_id = manager1.backup_file(sample_file)
        
        # 创建另一个管理器实例，应该加载已存在的元数据
        manager2 = FileBackupManager(backup_dir=temp_backup_dir)
        
        # 检查元数据是否被正确加载
        assert backup_id in manager2.metadata
        assert manager2.metadata[backup_id]["original_path"] == sample_file 


"""
Pruner Tools Module

This module provides command-line tools for managing and querying pruner-related data.
"""

# Import main functions for programmatic use
from .query_message_ids import (
    query_conversation_message_ids,
    list_all_conversations,
    export_all_configs
)

__all__ = [
    "query_conversation_message_ids",
    "list_all_conversations", 
    "export_all_configs"
] 


from autocoder.utils import get_last_yaml_file
import os
import uuid
import byzerllm
from autocoder.common import AutoCoderArgs
from autocoder.common.autocoderargs_parser import AutoCoderArgsParser
from typing import List, Optional
import yaml
from autocoder.auto_coder import AutoCoderArgs, load_include_files, Template
from autocoder.common import git_utils
import hashlib

# 该文件要给 chat,rag, web 等上层交互层使用


def convert_yaml_to_config(yaml_file: str):

    args = AutoCoderArgs()
    with open(yaml_file, "r",encoding="utf-8") as f:
        config = yaml.safe_load(f)
        config = load_include_files(config, yaml_file)
        for key, value in config.items():
            if key != "file":  # 排除 --file 参数本身
                # key: ENV {{VARIABLE_NAME}}
                if isinstance(value, str) and value.startswith("ENV"):
                    template = Template(value.removeprefix("ENV").strip())
                    value = template.render(os.environ)
                setattr(args, key, value)
    return args


def convert_yaml_config_to_str(yaml_config):
    yaml_content = yaml.safe_dump(
        yaml_config,
        allow_unicode=True,
        default_flow_style=False,
        default_style=None,
    )
    return yaml_content


def get_llm_friendly_package_docs(memory,
                                  package_name: Optional[str] = None, return_paths: bool = False
                                  ) -> List[str]:
    """
    Legacy function for backward compatibility.
    Use LLMFriendlyPackageManager class for new code.
    """
    from autocoder.common.llm_friendly_package import LLMFriendlyPackageManager
    
    project_root = os.getcwd()
    base_persist_dir = os.path.join(project_root, ".auto-coder", "plugins", "chat-auto-coder")
    
    manager = LLMFriendlyPackageManager(
        project_root=project_root,
        base_persist_dir=base_persist_dir
    )
    return manager.get_docs(package_name, return_paths)


def convert_config_value(key, value):
    # 定义需要使用 token 解析的字段
    token_fields = {
        'conversation_prune_safe_zone_tokens',
        'context_prune_safe_zone_tokens',
        'context_prune_sliding_window_size',
        'context_prune_sliding_window_overlap',
        'rag_params_max_tokens',
        'rag_context_window_limit',
        'rag_duckdb_vector_dim',
        'rag_duckdb_query_top_k',
        'rag_emb_dim',
        'rag_emb_text_size',
        'hybrid_index_max_output_tokens',
        'data_cells_max_num',
    }
    
    field_info = AutoCoderArgs.model_fields.get(key)
    if field_info:
        # 对于需要 token 解析的字段，使用 AutoCoderArgsParser
        if key in token_fields:
            try:
                parser = AutoCoderArgsParser()
                return parser.parse_token_field(key, value)
            except Exception as e:
                print(f"Warning: Failed to parse token field '{key}' with value '{value}': {e}")
                # 如果解析失败，fallback 到原有逻辑
                pass
        
        # 原有的类型转换逻辑
        if isinstance(value, str) and value.lower() in ["true", "false"]:
            return value.lower() == "true"
        elif "float" in str(field_info.annotation):
            return float(value)    
        elif "int" in str(field_info.annotation):
            return int(value)        
        else:
            return value
    else:
        print(f"Invalid configuration key: {key}")
        return None


def get_llm(memory, model:Optional[str]=None):
    latest_yaml_file = get_last_yaml_file("actions")

    conf = memory.get("conf", {})
    current_files = memory["current_files"]["files"]
    execute_file = None

    if latest_yaml_file:
        try:
            execute_file = os.path.join("actions", latest_yaml_file)
            yaml_config = {
                "include_file": ["./base/base.yml"],
                "auto_merge": conf.get("auto_merge", "editblock"),
                "human_as_model": conf.get("human_as_model", "false") == "true",
                "skip_build_index": conf.get("skip_build_index", "true") == "true",
                "skip_confirm": conf.get("skip_confirm", "true") == "true",
                "silence": conf.get("silence", "true") == "true",
                "include_project_structure": conf.get("include_project_structure", "false")
                == "true",
            }
            for key, value in conf.items():
                converted_value = convert_config_value(key, value)
                if converted_value is not None:
                    yaml_config[key] = converted_value

            yaml_config["urls"] = current_files + get_llm_friendly_package_docs(
                memory=memory,
                return_paths=True
            )

            # 临时保存yaml文件，然后读取yaml文件，转换为args
            temp_yaml = os.path.join("actions", f"{uuid.uuid4()}.yml")
            try:
                with open(temp_yaml, "w",encoding="utf-8") as f:
                    f.write(convert_yaml_config_to_str(
                        yaml_config=yaml_config))
                args = convert_yaml_to_config(temp_yaml)
            finally:
                if os.path.exists(temp_yaml):
                    os.remove(temp_yaml)

            llm = byzerllm.ByzerLLM.from_default_model(model or
                                                       args.code_model or args.model)
            return llm
        except Exception as e:
            print(f"Failed to commit: {e}")
            if execute_file:
                os.remove(execute_file)
            return None



"""
ReadImageToolResolver - 图片识别工具解析器

使用视觉模型（vl_model）分析图片内容，支持多图片路径（逗号分割）。
"""

import base64
import mimetypes
from pathlib import Path
from typing import Optional, List

import byzerllm
from loguru import logger

from autocoder.common import AutoCoderArgs
from autocoder.common.v2.agent.agentic_edit_tools.base_tool_resolver import (
    BaseToolResolver,
)
from autocoder.common.v2.agent.agentic_edit_types import ReadImageTool, ToolResult
from autocoder.utils.llms import get_single_llm
import typing

if typing.TYPE_CHECKING:
    from autocoder.common.v2.agent.agentic_edit import AgenticEdit


class ReadImageToolResolver(BaseToolResolver):
    """图片识别工具解析器

    使用 args.vl_model 指定的视觉模型读取和分析图片内容。
    支持多个图片路径（用逗号分割）。
    """

    # 支持的图片格式
    SUPPORTED_EXTENSIONS = {".jpg", ".jpeg", ".png", ".gif", ".webp", ".bmp"}

    def __init__(
        self, agent: Optional["AgenticEdit"], tool: ReadImageTool, args: AutoCoderArgs
    ):
        super().__init__(agent, tool, args)
        self.tool: ReadImageTool = tool

    def _get_vl_llm(self) -> Optional[byzerllm.ByzerLLM]:
        """获取视觉模型 LLM 实例"""
        vl_model = self.args.vl_model
        if not vl_model:
            return None

        try:
            return get_single_llm(vl_model, product_mode=self.args.product_mode)
        except Exception as e:
            logger.warning(f"Failed to get vl_model '{vl_model}': {e}")
            return None

    def _validate_image_path(self, path: str) -> tuple[bool, str]:
        """验证图片路径

        Returns:
            tuple[bool, str]: (是否有效, 错误信息或空字符串)
        """
        if not path:
            return False, "Empty path"

        image_path = Path(path)

        # 检查文件是否存在
        if not image_path.exists():
            return False, f"File not found: {path}"

        # 检查是否是文件
        if not image_path.is_file():
            return False, f"Not a file: {path}"

        # 检查文件扩展名
        ext = image_path.suffix.lower()
        if ext not in self.SUPPORTED_EXTENSIONS:
            return (
                False,
                f"Unsupported image format '{ext}'. Supported: {', '.join(self.SUPPORTED_EXTENSIONS)}",
            )

        return True, ""

    def _load_image_as_data_uri(self, image_path: str) -> str:
        """将本地图片加载为 data URI 格式

        Args:
            image_path: 图片文件路径

        Returns:
            data URI 格式的字符串，如 "data:image/png;base64,..."
        """
        mime_type = mimetypes.guess_type(image_path)[0] or "image/jpeg"
        with open(image_path, "rb") as f:
            base64_data = base64.b64encode(f.read()).decode("utf-8")
        return f"data:{mime_type};base64,{base64_data}"

    def _read_images(
        self,
        vl_llm: byzerllm.ByzerLLM,
        image_paths: List[str],
        query: str,
    ) -> tuple[bool, str]:
        """读取并分析图片（支持单张或多张）

        使用 OpenAI Vision 标准格式。

        Args:
            vl_llm: 视觉模型 LLM 实例
            image_paths: 图片文件路径列表
            query: 用户查询

        Returns:
            tuple[bool, str]: (是否成功, 分析结果或错误信息)
        """
        try:
            # 构建 content 列表，使用 OpenAI Vision 标准格式
            content = [{"type": "text", "text": query}]

            # 添加所有图片
            for image_path in image_paths:
                image_data_uri = self._load_image_as_data_uri(image_path)
                content.append(
                    {"type": "image_url", "image_url": {"url": image_data_uri}}
                )

            # 构建对话消息
            conversations = [{"role": "user", "content": content}]

            # 调用视觉模型
            response = vl_llm.chat_oai(conversations=conversations)

            if response and len(response) > 0:
                return True, response[0].output
            else:
                return False, "No response from vision model"

        except Exception as e:
            logger.exception(f"Error analyzing images: {image_paths}")
            return False, f"Error analyzing images: {e}"

    def resolve(self) -> ToolResult:
        """解析并执行图片识别请求"""
        # 检查 vl_model 配置
        if not self.args.vl_model:
            return ToolResult(
                success=False,
                message="Vision model not configured. Please set 'vl_model' in args to use read_image tool.",
            )

        # 获取视觉模型实例
        vl_llm = self._get_vl_llm()
        if not vl_llm:
            return ToolResult(
                success=False,
                message=f"Failed to initialize vision model '{self.args.vl_model}'.",
            )

        # 解析图片路径（支持逗号分割的多个路径）
        raw_path = self.tool.path or ""
        paths = [p.strip() for p in raw_path.split(",") if p.strip()]

        if not paths:
            return ToolResult(success=False, message="'path' is required")

        query = self.tool.query or "请描述这张图片的内容"

        # 验证所有路径
        valid_paths: List[str] = []
        failures: List[str] = []

        for i, image_path in enumerate(paths):
            valid, error_msg = self._validate_image_path(image_path)
            if valid:
                valid_paths.append(image_path)
            else:
                failures.append(f"[{i + 1}] {error_msg}")

        if not valid_paths:
            return ToolResult(
                success=False,
                message="No valid images found. " + "; ".join(failures),
            )

        # 一次调用分析所有有效图片
        success, result = self._read_images(vl_llm, valid_paths, query)

        if not success:
            return ToolResult(success=False, message=result)

        # 构建返回结果
        summary = f"Analyzed {len(valid_paths)} image(s): {', '.join(valid_paths)}"
        if failures:
            summary += f" | skipped: {'; '.join(failures)}"

        return ToolResult(
            success=True,
            message=summary,
            content=result,
        )
